# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


from . import templates
import json


def prepare_unit(ctx):
    """
    Prepare the unit for lifecycle events.
    """
    # Tell the render engine about our templates
    ctx.render.register_package(templates)

    # Register for events
    ctx.events.verify_config.register("dns_v1", handle_verify_config)
    ctx.events.ready_apply.register("dns_v1", handle_ready_apply)
    ctx.events.check_health.register("dns_v1", handle_check_health)


def handle_verify_config(evt):
    """
    Handle a verify_config event.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.report_issue to report any found issues.

    Args:
        evt: A VerifyConfigEvent instance.
    """


def _ip_rules_get(evt):
    try:
        output = str(evt.sh.ip(*"-j rule list ipproto udp dport 53".split()))
        rules = json.loads(output)
        return rules
    except Exception:
        return []


def _get_configured_dns_servers(evt):
    dns_servers = []

    rules = _ip_rules_get(evt)
    for rule in rules:
        dns_servers.append((rule["dst"], rule["table"]))

    return dns_servers


def handle_dns_vrf_rule_config(evt, dns_conf):
    old = _get_configured_dns_servers(evt)
    new = []

    for vrf in dns_conf:
        dns_servers = dns_conf[vrf]["server"]
        for server in dns_servers:
            new.append((server, vrf))
    stale = []
    for old_server, old_vrf in old:
        found = False
        for new_server, new_vrf in new:
            if (old_server == new_server) and (old_vrf == new_vrf):
                found = True
                break
        if not found:
            stale.append((old_server, old_vrf))

    if (len(stale) > 0) or (len(new) > 0):
        evt.render(templates, "ip_rule_mod.sh", del_servers=stale,
                   add_servers=new)
        evt.schedule_run_script("ip_rule_mod.sh")


def handle_ready_apply(evt):
    """
    Prepare to apply the config.  Generate config files and scripts.  Also,
    look for issues in the config that would prevent us from applying it.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.render to render a template with the given variables.
    Use evt.stage_file to prepare a rendered file to be copied over to its
        ultimate destination.

    Args:
        evt: A ReadyApplyEvent instance.
    """

    # check eth address is type dhcp, if so, no need to
    # update nameserver in resolv.conf
    ifaces = evt.config_v1.getInterfaces()
    for iface in ifaces.values():
        if iface['type'] == 'eth' and 'dhcp' in iface["ip"]["address"]:
            return

    # Get the lists of DNS server IPs and update resolv.conf
    dns_configs = evt.config_v1.getDnss()
    handle_dns_vrf_rule_config(evt, dns_configs)
    evt.render(templates, "resolv.conf",
               dns=dns_configs)

    # Todo handle mgmt ip, is it not required in resolv.conf?
    # host_entry = evt.config_v1.getHostEntries()
    # exist = bool(host_entry)
    # if exist is True:
    #   mgmt_server = host_entry.get('mgmt-server')
    #   ip_addr = mgmt_server.get('ip-address')
    #   ip = ip_addr.keys()

    # updating config file
    evt.stage_file("resolv.conf", "/etc/resolv.conf")


def handle_check_health(evt):
    """
    Handle a check_health event.

    Use evt.report_error to report any found errors.
    Use evt.report_warning to report any found warnings.

    Warnings *should* clear on their own, given some time, usually indicating
    that a component is still "coming up."  Errors *will not* clear on their
    own.

    Args:
        evt: A CheckHealthEvent instance.
    """
