# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


from . import templates


def prepare_unit(ctx):
    """
    Prepare the unit for lifecycle events.
    """
    # Tell the render engine about our templates
    ctx.render.register_package(templates)

    # Register for events
    ctx.events.verify_config.register("ntp_v1", handle_verify_config)
    ctx.events.ready_apply.register("ntp_v1", handle_ready_apply)
    ctx.events.check_health.register("ntp_v1", handle_check_health)


def handle_verify_config(evt):
    """
    Handle a verify_config event.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.report_issue to report any found issues.

    Args:
        evt: A VerifyConfigEvent instance.
    """


def get_running_vrfs(sh):
    """
    Return the list of VRFs on which NTP is running.
    """
    try:
        out = sh.systemctl("list-units", "ntp@*", "--no-legend")
    except Exception:
        # Don't panic.
        return {}

    vrfs = {}
    for line in out.splitlines():
        # Extract the vrf name
        tokens = line.split()
        service, _, _ = tokens[0].rpartition(".service")
        if service:
            vrfs[service[4:]] = service

    return vrfs


def handle_stale_config(evt, ntps):
    """
    Check the existing NTP services and stop/disable any that we no longer
    need.
    """
    vrfs = get_running_vrfs(evt.sh)
    for vrf, service in vrfs.items():
        # If this vrf isn't in our list of running instance, then stop/disable.
        if vrf not in ntps:
            evt.schedule_stop(service)
            evt.schedule_disable(service)


def handle_ready_apply(evt):
    """
    Prepare to apply the config.  Generate config files and scripts.  Also,
    look for issues in the config that would prevent us from applying it.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.render to render a template with the given variables.
    Use evt.stage_file to prepare a rendered file to be copied over to its
        ultimate destination.

    Args:
        evt: A ReadyApplyEvent instance.
    """
    ntps = evt.config_v1.getNtps()
    handle_stale_config(evt, ntps)

    # We support at most 1 instance.
    if len(ntps) == 0:
        evt.render(templates, "ntp.conf", enabled=False)
        service = None
    else:
        vrf, ntp = ntps.popitem()
        evt.render(templates, "ntp.conf", enabled=True, config=ntp, vrf=vrf)
        service = "ntp@" + vrf

    if evt.stage_file("ntp.conf", "/etc/ntp.conf"):
        if service:
            evt.schedule_enable(service)
            evt.schedule_reload_or_restart(service)


def handle_check_health(evt):
    """
    Handle a check_health event.

    Use evt.report_error to report any found errors.
    Use evt.report_warning to report any found warnings.

    Warnings *should* clear on their own, given some time, usually indicating
    that a component is still "coming up."  Errors *will not* clear on their
    own.

    Args:
        evt: A CheckHealthEvent instance.
    """
    # Make sure the service is running
    # evt.check_is_active('ntp@mgmt')

    # And responding to us
    # TODO CUE-3802: Commenting this out for now until we can figure
    # out why sh.ntqp("-c", "sysnfo") doesn't run output anything when run
    # by CUE.
    # if "system peer" not in str(evt.sh.ntpq("-c", "sysinfo")):
    #     evt.report_warning("NTP daemon is not responding")
